use std::{
    collections::HashMap,
    ops::{Deref, DerefMut},
    sync::Arc,
};

use chat::{Body, Message};
use connection::{write::WriteMessage, SupervisorSender};
use jabber::JID;
use presence::{Offline, Online, Presence};
use roster::Contact;
use sqlx::SqlitePool;
use stanza::client::{
    iq::{self, Iq, IqType},
    Stanza,
};
use tokio::{
    sync::{mpsc, oneshot, Mutex},
    task::JoinSet,
};
use uuid::Uuid;

use crate::connection::write::WriteHandle;
use crate::connection::{SupervisorCommand, SupervisorHandle};
use crate::error::Error;

mod chat;
mod connection;
mod error;
mod presence;
mod roster;
mod user;

pub struct Luz {
    receiver: mpsc::Receiver<CommandMessage>,
    jid: Arc<Mutex<JID>>,
    // TODO: use a dyn passwordprovider trait to avoid storing password in memory
    password: Arc<String>,
    connected: Arc<Mutex<Option<(WriteHandle, SupervisorHandle)>>>,
    pending_iqs: Arc<Mutex<HashMap<String, oneshot::Sender<Result<Stanza, Error>>>>>,
    db: SqlitePool,
    sender: mpsc::Sender<UpdateMessage>,
    /// if connection was shut down due to e.g. server shutdown, supervisor must be able to mark client as disconnected
    connection_supervisor_shutdown: oneshot::Receiver<()>,
    // TODO: will need to have an auto reconnect state as well (e.g. in case server shut down, to try and reconnect later)
    // TODO: will grow forever at this point, maybe not required as tasks will naturally shut down anyway?
    tasks: JoinSet<()>,
}

impl Luz {
    fn new(
        receiver: mpsc::Receiver<CommandMessage>,
        jid: Arc<Mutex<JID>>,
        password: String,
        connected: Arc<Mutex<Option<(WriteHandle, SupervisorHandle)>>>,
        connection_supervisor_shutdown: oneshot::Receiver<()>,
        db: SqlitePool,
        sender: mpsc::Sender<UpdateMessage>,
    ) -> Self {
        Self {
            jid,
            password: Arc::new(password),
            connected,
            db,
            receiver,
            sender,
            tasks: JoinSet::new(),
            connection_supervisor_shutdown,
            pending_iqs: Arc::new(Mutex::new(HashMap::new())),
        }
    }

    async fn run(mut self) {
        loop {
            tokio::select! {
                // this is okay, as when created the supervisor (and connection) doesn't exist, but a bit messy
                _ = &mut self.connection_supervisor_shutdown => {
                    *self.connected.lock().await = None
                }
                Some(msg) = self.receiver.recv() => {
                    // TODO: consider separating disconnect/connect and commands apart from commandmessage
                    // TODO: dispatch commands separate tasks
                    match msg {
                        CommandMessage::Connect => {
                            let mut connection_lock = self.connected.lock().await;
                            match connection_lock.as_ref() {
                                Some(_) => {
                                    self.sender
                                        .send(UpdateMessage::Error(Error::AlreadyConnected))
                                        .await;
                                }
                                None => {
                                    let mut jid = self.jid.lock().await;
                                    let mut domain = jid.domainpart.clone();
                                    // TODO: check what happens upon reconnection with same resource (this is probably what one wants to do and why jid should be mutated from a bare jid to one with a resource)
                                    let streams_result =
                                        jabber::connect_and_login(&mut jid, &*self.password, &mut domain)
                                            .await;
                                    match streams_result {
                                        Ok(s) => {
                                            let (shutdown_send, shutdown_recv) = oneshot::channel::<()>();
                                            let (writer, supervisor) = SupervisorHandle::new(
                                                s,
                                                self.sender.clone(),
                                                self.db.clone(),
                                                shutdown_send,
                                                self.jid.clone(),
                                                self.password.clone(),
                                                self.pending_iqs.clone(),
                                            );
                                            self.connection_supervisor_shutdown = shutdown_recv;
                                            *connection_lock = Some((writer, supervisor));
                                            self.sender
                                                .send(UpdateMessage::Connected)
                                                .await;
                                        }
                                        Err(e) => {
                                            self.sender.send(UpdateMessage::Error(e.into()));
                                        }
                                    }
                                }
                            };
                        }
                        CommandMessage::Disconnect => match self.connected.lock().await.as_mut() {
                            None => {
                                self.sender
                                    .send(UpdateMessage::Error(Error::AlreadyDisconnected))
                                    .await;
                            }
                            mut c => {
                                if let Some((_write_handle, supervisor_handle)) = c.take() {
                                    let _ = supervisor_handle.send(SupervisorCommand::Disconnect).await;
                                } else {
                                    unreachable!()
                                };
                            }
                        },
                        _ => {
                            match self.connected.lock().await.as_ref() {
                                Some((w, s)) => self.tasks.spawn(msg.handle_online(
                                    w.clone(),
                                    s.sender(),
                                    self.jid.clone(),
                                    self.db.clone(),
                                    self.sender.clone(),
                                    self.pending_iqs.clone()
                                )),
                                None => self.tasks.spawn(msg.handle_offline(
                                    self.jid.clone(),
                                    self.db.clone(),
                                    self.sender.clone(),
                                )),
                            };
                        }
                    }
                },
                else => break,
            }
        }
    }
}

impl CommandMessage {
    pub async fn handle_offline(
        mut self,
        jid: Arc<Mutex<JID>>,
        db: SqlitePool,
        sender: mpsc::Sender<UpdateMessage>,
    ) {
        todo!()
    }

    pub async fn handle_online(
        mut self,
        write_handle: WriteHandle,
        supervisor_control: SupervisorSender,
        // TODO: jid could lose resource by the end
        jid: Arc<Mutex<JID>>,
        db: SqlitePool,
        sender: mpsc::Sender<UpdateMessage>,
        pending_iqs: Arc<Mutex<HashMap<String, oneshot::Sender<Result<Stanza, Error>>>>>,
    ) {
        match self {
            CommandMessage::Connect => unreachable!(),
            CommandMessage::Disconnect => unreachable!(),
            CommandMessage::GetRoster => {
                // TODO: jid resource should probably be stored within the connection
                let owned_jid: JID;
                {
                    owned_jid = jid.lock().await.clone();
                }
                let stanza = Stanza::Iq(Iq {
                    from: Some(owned_jid),
                    id: "getting-roster".to_string(),
                    to: None,
                    r#type: IqType::Get,
                    lang: None,
                    query: Some(iq::Query::Roster(stanza::roster::Query {
                        ver: None,
                        items: Vec::new(),
                    })),
                    errors: Vec::new(),
                });
                let (send, recv) = oneshot::channel();
                let _ = write_handle
                    .send(WriteMessage {
                        stanza,
                        respond_to: send,
                    })
                    .await;
                match recv.await {
                    Ok(Ok(())) => println!("roster request sent"),
                    e => println!("error: {:?}", e),
                };
            }
            CommandMessage::SendMessage(jid, _) => todo!(),
        }
    }
}

// TODO: separate sender and receiver, store handle to Luz process to ensure dropping
// #[derive(Clone)]
pub struct LuzHandle {
    sender: mpsc::Sender<CommandMessage>,
}

impl Deref for LuzHandle {
    type Target = mpsc::Sender<CommandMessage>;

    fn deref(&self) -> &Self::Target {
        &self.sender
    }
}

impl DerefMut for LuzHandle {
    fn deref_mut(&mut self) -> &mut Self::Target {
        &mut self.sender
    }
}

impl LuzHandle {
    pub fn new(
        jid: JID,
        password: String,
        db: SqlitePool,
    ) -> (Self, mpsc::Receiver<UpdateMessage>) {
        let (command_sender, command_receiver) = mpsc::channel(20);
        let (update_sender, update_receiver) = mpsc::channel(20);
        // might be bad, first supervisor shutdown notification oneshot is never used (disgusting)
        let (sup_send, sup_recv) = oneshot::channel();

        let actor = Luz::new(
            command_receiver,
            Arc::new(Mutex::new(jid)),
            password,
            Arc::new(Mutex::new(None)),
            sup_recv,
            db,
            update_sender,
        );
        tokio::spawn(async move { actor.run().await });

        (
            Self {
                sender: command_sender,
            },
            update_receiver,
        )
    }
}

pub enum CommandMessage {
    // TODO: login invisible xep-0186
    /// connect to XMPP chat server. gets roster and publishes initial presence.
    Connect,
    /// disconnect from XMPP chat server, sending unavailable presence then closing stream.
    Disconnect,
    /// get the roster. if offline, retreive cached version from database. should be stored in application memory
    GetRoster,
    // add a contact to your roster, with a status of none, no subscriptions
    AddContact(JID),
    /// send a friend request i.e. a subscription request with a subscription pre-approval. if not already added to roster server adds to roster.
    BuddyRequest(JID),
    /// send a subscription request, without pre-approval. if not already added to roster server adds to roster.
    SubscriptionRequest(JID),
    /// accept a friend request by accepting a pending subscription and sending a subscription request back. if not already added to roster adds to roster.
    AcceptBuddyRequest(JID),
    /// accept a pending subscription and doesn't send a subscription request back. if not already added to roster adds to roster.
    AcceptSubscriptionRequest(JID),
    /// unsubscribe to a contact, but don't remove their subscription.
    UnsubscribeFromContact(JID),
    /// stop a contact from being subscribed, but stay subscribed to the contact.
    UnsubscribeContact(JID),
    /// remove subscriptions to and from contact, but keep in roster.
    UnfriendContact(JID),
    /// remove a contact from the contact list. will remove subscriptions if not already done then delete contact from roster.
    DeleteContact(JID),
    /// set online status
    SendStatus(Online),
    SendOffline(Offline),
    /// send a directed presence (usually to a non-contact).
    // TODO: should probably make it so people can add non-contact auto presence sharing in the client.
    SendDirectedPresence {
        to: JID,
        presence: Presence,
    },
    SendMessage {
        id: Uuid,
        to: JID,
        body: Body,
    },
}

#[derive(Debug)]
pub enum UpdateMessage {
    Error(Error),
    Connected(Online),
    Disconnected(Offline),
    /// full roster (replace full app roster state with this)
    Roster(Vec<Contact>),
    /// roster update (only update app roster state)
    RosterPush(Contact),
    Presence {
        from: JID,
        presence: Presence,
    },
    MessageDispatched(Uuid),
    Message {
        from: JID,
        message: Message,
    },
}
